@@ -3,7 +3,7 @@
 #
 #  Smooth the marks of a point pattern
 # 
-#  $Revision: 1.80 $  $Date: 2022/05/21 08:53:38 $
+#  $Revision: 1.103 $  $Date: 2025/05/13 07:01:00 $
 #
 
 Smooth <- function(X, ...) {
@@ -16,44 +16,66 @@ Smooth.solist <- function(X, ...) {
 
 Smooth.ppp <- function(X, sigma=NULL, ...,
                        weights=rep(1, npoints(X)), at="pixels",
+                       leaveoneout=TRUE,
                        adjust=1, varcov=NULL, 
                        edge=TRUE, diggle=FALSE, 
                        kernel="gaussian",
                        scalekernel=is.character(kernel),
-                       geometric=FALSE) {
+                       se=FALSE,
+                       loctype=c("random", "fixed"),
+                       wtype=c("multiplicity", "importance"),
+                       geometric=FALSE,
+                       shrink=0, shrinktype=c("mean", "median")) {
   verifyclass(X, "ppp")
   if(!is.marked(X, dfok=TRUE, na.action="fatal"))
     stop("X should be a marked point pattern", call.=FALSE)
+  nX <- npoints(X)
   X <- coerce.marks.numeric(X)
-  if(!all(is.finite(as.matrix(marks(X)))))
+  marx <- marks(X)
+  if(!all(is.finite(as.matrix(marx))))
     stop("Some mark values are Inf, NaN or NA", call.=FALSE)
+  univariate <- is.null(dim(marx))
+  nc <- if(univariate) 1 else ncol(marx)
+  
+  ## options
   at <- pickoption("output location type", at,
                    c(pixels="pixels",
                      points="points"))
+  loctype <- match.arg(loctype)
+  wtype <- match.arg(wtype)
+
   ## trivial case
-  if(npoints(X) == 0) {
+  if(nX == 0) {
     cn <- colnames(marks(X))
-    nc <- length(cn)
     switch(at,
            points = {
-             result <- if(nc == 0) numeric(0) else
-             matrix(, 0, nc, dimnames=list(NULL, cn))
+             Estimate <- if(univariate) numeric(0) else
+                       matrix(, 0, nc, dimnames=list(NULL, cn))
+             result <- if(!se) Estimate else
+                       cbind(estimate=Estimate, SE=Estimate)
            },
            pixels = {
-             result <- as.im(NA_real_, Window(X))
-             if(nc) {
-               result <- as.solist(rep(list(result), nc))
-               names(result) <- cn
+             Estimate <- as.im(NA_real_, Window(X))
+             if(!univariate) {
+               Estimate <- as.solist(rep(list(Estimate), nc))
+               names(Estimate) <- cn
              }
-           })
+             result <- if(!se) {
+                         Estimate
+                       } else if(univariate) {
+                         solist(estimate=Estimate, SE=Estimate)
+                       } else {
+                         list(estimate=Estimate, SE=Estimate)
+                       }
+             })
     return(result)
   }
 
   ## ensure weights are numeric
-  if(weightsgiven <- !missing(weights) && !is.null(weights)) {
+  if(weighted <- !missing(weights) && !is.null(weights)) {
     pa <- parent.frame()
     weights <- pointweights(X, weights=weights, parent=pa)
-    weightsgiven <- !is.null(weights)
+    weighted <- !is.null(weights)
   } else weights <- NULL
 
   ## geometric mean smoothing
@@ -61,6 +83,7 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
     return(ExpSmoothLog(X, sigma=sigma, ..., at=at,
                         adjust=adjust, varcov=varcov,
                         kernel=kernel, scalekernel=scalekernel,
+                        se=se, loctype=loctype, wtype=wtype,
                         weights=weights, edge=edge, diggle=diggle))
 
   ## determine smoothing parameters
@@ -74,39 +97,123 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
     adjust <- 1
   } 
   
-  ## infinite bandwidth
+  ## ............ infinite bandwidth ...............................
   if(bandwidth.is.infinite(sigma)) {
     #' uniform estimate
-    nX <- npoints(X)
-    if(is.null(weights)) weights <- rep(1, nX)
-    marx <- marks(X)
-    single <- is.null(dim(marx))
+    if(!weighted) weights <- rep(1, nX)
     wtmark <- weights * marx 
     totwt <- sum(weights)
-    totwtmark <- if(single) sum(wtmark) else colSums(wtmark)
+    totwtmark <- if(univariate) sum(wtmark) else colSums(wtmark)
     W <- Window(X)
     switch(at,
            pixels = {
-             result <- solapply(totwtmark/totwt, as.im, W=W, ...)
-             names(result) <- colnames(marx)
-             if(single) result <- result[[1L]]
+             Estimate <- solapply(totwtmark/totwt, as.im, W=W, ...)
+             names(Estimate) <- colnames(marx)
+             if(univariate) Estimate <- Estimate[[1L]]
            },
            points = {
              denominator <- rep(totwt, nX)
              numerator <- rep(totwtmark, each=nX)
-             if(!single) numerator <- matrix(numerator, nrow=nX)
-             leaveoneout <- resolve.1.default(list(leaveoneout=TRUE), list(...))
+             if(!univariate) numerator <- matrix(numerator, nrow=nX)
              if(leaveoneout) {
                numerator <- numerator - wtmark
                denominator <- denominator - weights
              }
-             result <- numerator/denominator
-             if(!single)
-               colnames(result) <- colnames(marx)
+             Estimate <- numerator/denominator
+             if(!univariate)
+               colnames(Estimate) <- colnames(marx)
            })
+    if(!se) {
+      result <- Estimate
+    } else {
+      ## calculate standard error (constant value)
+      if(univariate) {
+        V <- if(!weighted) var(marx) else
+             switch(loctype,
+                    fixed = {
+                      switch(wtype,
+                             multiplicity = weighted.var(marx, weights),
+                             importance = var(marx * weights))
+                    },
+                    random = {
+                      switch(wtype,
+                             multiplicity = VarOfWtdMean(marx, weights),
+                             importance = VarOfWtdMean(marx, weights^2))
+                    })
+        SE <- sqrt(V) # single value
+      } else {
+        V <- if(!weighted) sapply(marx, var) else
+             switch(loctype,
+                    fixed = {
+                      switch(wtype,
+                             multiplicity = sapply(marx, weighted.var,
+                                                   wt=weights),
+                             importance = sapply(marx * weights, var))
+                    },
+                    random = {
+                      switch(wtype,
+                             multiplicity = sapply(marx, VarOfWtdMean,
+                                                   weights=weights),
+                             importance = sapply(marx, VarOfWtdMean,
+                                                 weights=weights^2))
+                    })
+        SE <- sqrt(V) # vector
+      }
+      ## replicate constant value
+      switch(at,
+             pixels = {
+               if(univariate) {
+                 SE <- as.im(SE, W=W, ...) # constant image
+                 result <- solist(estimate=Estimate, SE=SE)
+               } else {
+                 SE <- solapply(SE, as.im, ...) # list of images
+                 result <- list(estimate=Estimate, SE=SE)
+               }
+             },
+             points = {
+               if(univariate) {
+                 SE <- rep(SE, nX)
+                 result <- cbind(estimate=Estimate, SE=SE)
+               } else {
+                 SE <- matrix(SE[col(marx)], nrow=nX, ncol=ncol(marx))
+                 colnames(SE) <- colnames(marx)
+                 result <- cbind(estimate=Estimate, SE=SE)
+               }
+             })
+    }
     return(result)
   }
 
+  ## .................. finite bandwidth .............................
+
+  ## shrinkage
+  nc <- if(is.null(dim(marx))) 1 else ncol(marx)
+  if(shrinking <- (!missing(shrink) && (length(shrink) > 0))) {
+    check.nvector(shrink, nc, 
+                  things="columns of marks", oneok=TRUE, vname="shrink")
+    stopifnot(all(shrink >= 0))
+    if(length(shrink) == 1 && nc > 1)
+      shrink <- rep(shrink, nc)
+    ## Rescale by Kernel at 0
+    K0 <- evaluate2Dkernel(kernel, 0, 0, sigma=sigma, varcov=varcov)
+    shrinkdenom <- shrink * K0
+    ## Numerator shrinkage constant involves mean or median mark 'ybar'
+    shrinktype <- match.arg(shrinktype)
+    marxmat <- if(univariate) matrix(marx, ncol=1) else marx
+    if(weighted) {
+      ybar <- switch(shrinktype,
+                     mean = apply(marxmat, 2, weighted.mean, w=weights),
+                     median = apply(marxmat, 2, weighted.median, w=weights))
+    } else {
+      ybar <- switch(shrinktype,
+                     mean = colMeans(marxmat, na.rm=TRUE),
+                     median = apply(marxmat, 2, median, na.rm=TRUE))
+    }
+    shrinknumer <- shrinkdenom * ybar
+  } else {
+    shrinknumer <- shrinkdenom <- rep(0, nc)
+  }
+  
   ## Diggle's edge correction?
   if(diggle && !edge) warning("Option diggle=TRUE overridden by edge=FALSE")
   diggle <- diggle && edge
@@ -115,39 +222,149 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
   cutoff <- cutoff2Dkernel(kernel, sigma=sigma, varcov=varcov,
                            scalekernel=scalekernel, adjust=adjust, ...,
                            fatal=TRUE)
-  ## 
+  
+  ## ...................  bandwidth close to zero .....................
   if(cutoff < minnndist(X)) {
     # very small bandwidth
-    leaveoneout <- resolve.1.default("leaveoneout",
-                                     list(...), list(leaveoneout=TRUE))
     if(!leaveoneout && at=="points") {
       warning(paste("Bandwidth is close to zero:",
                     "original values returned"))
-      Y <- marks(X)
+      Estimate <- marks(X)
     } else {
       warning(paste("Bandwidth is close to zero:",
                     "nearest-neighbour interpolation performed"))
-      Y <- nnmark(X, ..., k=1, at=at)
+      Estimate <- nnmark(X, ..., k=1, at=at)
     }
-    return(Y)
+    if(!se) {
+      result <- Estimate
+    } else {
+      SE <- 0 * Estimate
+      switch(at,
+             pixels = {
+               if(univariate) {
+                 result <- solist(estimate=Estimate, SE=SE)
+               } else {
+                 result <- list(estimate=Estimate, SE=SE)
+               }
+             },
+             points = {
+               result <- cbind(estimate=Estimate, SE=SE)
+             })
+    }
+    return(result)
   }
 
+  ## ................... bandwidth >> 0 .........................
+
+  if(se) {
+    ## ................... STANDARD ERROR CALCULATION ..............
+    ## This has to be done now because the subsequent code
+    ## fiddles with the weights.
+    weightspower <- if(!weighted) 1 else switch(wtype,
+                                                importance = weights^2,
+                                                multiplicity = weights)
+    if(diggle) {
+      ## Jones-Diggle correction weights e(x_i)
+      edgeim <- second.moment.calc(X, sigma, what="edge", ...,
+                                   varcov=varcov)
+      edgeX <- safelookup(edgeim, X, warn=FALSE)
+      invmassX <- 1/edgeX
+      invmassX[!is.finite(invmassX)] <- 0
+    } else {
+      invmassX <- 1
+    }
+    ## 
+    switch(loctype,
+           random = {
+             denom <- density(X, sigma=sigma, ...,
+                              weights=weights,
+                              edge=edge, diggle=diggle,
+                              at=at, leaveoneout=leaveoneout)
+             numer <- density(X, sigma=sigma, ...,
+                              weights=if(weighted) weights * marx else marx,
+                              edge=edge, diggle=diggle,
+                              at=at, leaveoneout=leaveoneout)
+             varNum <- density(X, sigma=sigma, ..., kerpow=2,
+                               weights=weightspower * marx^2 * invmassX^2,
+                               edge=FALSE, diggle=FALSE,
+                               at=at, leaveoneout=leaveoneout)
+             covND <- density(X, sigma=sigma, ..., kerpow=2,
+                              weights=weightspower * marx * invmassX^2,
+                              edge=FALSE, diggle=FALSE,
+                              at=at, leaveoneout=leaveoneout)
+             varDen <- density(X, sigma=sigma, ..., kerpow=2,
+                               weights=weightspower * invmassX^2,
+                               edge=FALSE, diggle=FALSE,
+                               at=at, leaveoneout=leaveoneout)
+             if(univariate || at == "points") {
+               Vest <- DeltaMethodVarOfRatio(numer, denom,
+                                          varNum, varDen, covND)
+             } else {
+               Vest <- mapply(DeltaMethodVarOfRatio,
+                           num=numer, varnum=varNum, covnumden=covND,
+                           MoreArgs = list(den=denom, varden=varDen),
+                           SIMPLIFY=FALSE)
+               Vest <- as.solist(Vest)
+             }
+           },
+           fixed = {
+             ## Use leave-one-out deviation
+             dev <- marks(X) - Smooth(X, sigma=sigma, ...,
+                                      weights=weights, edge=edge, diggle=diggle,
+                                      at="points", leaveoneout=TRUE)
+             if(!univariate)
+               dev <- asNumericMatrix(as.data.frame(dev))
+             ## calculate variance of numerator using leave-one-out estimates
+             dataweight <- dev^2
+             if(weighted)
+               dataweight <- dataweight * switch(wtype,
+                                                 importance=weights^2,
+                                                 multiplicity=weights)
+             if(edge && diggle) 
+               dataweight <- dataweight * invmassX^2
+
+             ## variance of numerator
+             Vnum <- density(unmark(X), sigma=sigma, kerpow=2,
+                             weights=dataweight,
+                             at=at, leaveoneout=leaveoneout,
+                             edge=FALSE, diggle=FALSE, # sic
+                             positive=TRUE)
+             
+             if(at == "points" && !univariate)
+               Vnum <- asNumericMatrix(as.data.frame(Vnum))
+    
+             ## rescale by denominator^2
+             Den <- density(unmark(X), sigma=sigma,
+                            weights=weights,
+                            edge=edge && diggle, diggle=diggle, # sic
+                            at=at, leaveoneout=leaveoneout, positive=TRUE)
+
+             Vest <- posify(if(at == "points") Vnum/Den^2 else
+                            imagelistOp(Vnum, Den^2, "/"))
+                     
+           })
+    SE <- if(is.solist(Vest)) solapply(Vest, sqrt) else sqrt(Vest)
+  }
+  
+  ##  ------------------------------------------------------------
+  ##  >>>>>>>>>>>. MAIN CALCULATION OF ESTIMATE <<<<<<<<<<<<<<<<<<
+  ##  ------------------------------------------------------------
+  
   if(diggle) {
     ## absorb Diggle edge correction into weights vector
     edg <- second.moment.calc(X, sigma, what="edge", ...,
                               varcov=varcov, adjust=adjust,
                               kernel=kernel, scalekernel=scalekernel)
     ei <- safelookup(edg, X, warn=FALSE)
-    weights <- if(weightsgiven) weights/ei else 1/ei
+    weights <- if(weighted) weights/ei else 1/ei
     weights[!is.finite(weights)] <- 0
-    weightsgiven <- TRUE
+    weighted <- TRUE
   }
   ## rescale weights to avoid numerical gremlins
-  if(weightsgiven && ((mw <- median(abs(weights))) > 0))
+  if(weighted && ((mw <- median(abs(weights))) > 0))
     weights <- weights/mw
 
   ## calculate...
-  marx <- marks(X)
   uhoh <- NULL
   if(!is.data.frame(marx)) {
     # ........ vector of marks ...................
@@ -174,14 +391,17 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
                          resolve.defaults(list(x=quote(X),
                                                values=quote(values), 
 					       weights=quote(weights),
+                                               leaveoneout=leaveoneout,
                                                sigma=sigma, varcov=varcov,
                                                kernel=kernel,
                                                scalekernel=scalekernel,
-                                               edge=FALSE),
+                                               edge=FALSE,
+                                               shrinknumer=shrinknumer,
+                                               shrinkdenom=shrinkdenom),
                                           list(...)))
              },
              pixels={
-               values.weights <- if(weightsgiven) values * weights else values
+               values.weights <- if(weighted) values * weights else values
                dont.complain.about(values.weights)
                numerator <-
                  do.call(density.ppp,
@@ -204,6 +424,10 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
                                                scalekernel=scalekernel,
                                                edge=FALSE),
                                           list(...)))
+               if(shrinking) {
+                 numerator   <- shrinknumer + numerator
+                 denominator <- shrinkdenom + denominator
+               }
                result <- eval.im(numerator/denominator)
                ## trap small values of denominator
                ## trap NaN and +/- Inf values of result, but not NA
@@ -211,6 +435,7 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
                nbg <- eval.im(is.infinite(result)
                               | is.nan(result)
                               | (denominator < eps))
+               uhoh <- attr(numerator, "warnings")
                if(any(as.matrix(nbg), na.rm=TRUE)) {
                  warning(paste("Numerical underflow detected:",
                                "sigma is probably too small"),
@@ -222,7 +447,6 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
                  nnvalues <- eval.im(values[whichnn])
                  result[nbg] <- nnvalues[nbg]
                }
-               uhoh <- attr(numerator, "warnings")
              })
     }
   } else {
@@ -246,6 +470,7 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
         do.call(density.ppp,
                 resolve.defaults(list(x=quote(X),
                                       at=at,
+                                      leaveoneout=leaveoneout,
                                       weights = quote(weights),
                                       sigma=sigma, varcov=varcov,
                                       kernel=kernel,
@@ -253,12 +478,13 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
                                       edge=FALSE),
                                  list(...)))
       ## compute numerator for each column of marks
-      marx.weights <- if(weightsgiven) marx * weights else marx
+      marx.weights <- if(weighted) marx * weights else marx
       dont.complain.about(marx.weights)
       numerators <-
         do.call(density.ppp,
                 resolve.defaults(list(x=quote(X),
                                       at=at,
+                                      leaveoneout=leaveoneout,
                                       weights = quote(marx.weights),
                                       sigma=sigma, varcov=varcov,
                                       kernel=kernel,
@@ -266,6 +492,26 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
                                       edge=FALSE),
                                  list(...)))
       uhoh <- attr(numerators, "warnings")
+      ## shrinkage estimator
+      if(shrinking) {
+        denominator <- shrinkdenom + denominator
+        switch(at,
+               points = {
+                 if(is.null(dim(numerators))) {
+                   numerators <- numerators + shrinknumer
+                 } else {
+                   ## add shrinknumer[j] to numerators[, j]
+                   numerators <- numerators + shrinknumer[col(numerators)]
+                 }
+               },
+               pixels = {
+                 ## add shrinknumer[j] to numerators[[j]]
+                 numerators <- as.imlist(mapply("+",
+                                                numerators,
+                                                as.list(shrinknumer),
+                                                SIMPLIFY=FALSE))
+               })
+      }
       ## calculate ratios
       switch(at,
              points={
@@ -347,7 +593,18 @@ Smooth.ppp <- function(X, sigma=NULL, ...,
   ## wrap up
   attr(result, "sigma") <- sigma
   attr(result, "varcov") <- varcov
-  attr(result, "warnings") <- uhoh
+  if(length(uhoh)) attr(result, "warnings") <- uhoh 
+  ## tack on standard errors?
+  if(se) {
+    result <- list(estimate=result, SE=SE)
+    switch(at,
+           points = {
+             result <- do.call(cbind, result)
+           },
+           pixels = {
+             if(univariate) result <- as.solist(result)
+           })
+  }
   return(result)
 }
 
@@ -358,7 +615,8 @@ smoothpointsEngine <- function(x, values, sigma, ...,
                                weights=NULL, varcov=NULL,
                                leaveoneout=TRUE,
                                sorted=FALSE, cutoff=NULL,
-                               debug=FALSE) {
+                               debug=FALSE,
+                               shrinknumer=0, shrinkdenom=0) {
   stopifnot(is.logical(leaveoneout))
 
   if(!is.null(dim(values)))
@@ -372,6 +630,8 @@ smoothpointsEngine <- function(x, values, sigma, ...,
     return(result)
   }
 
+  shrinking <- (shrinkdenom != 0)
+
   validate2Dkernel(kernel)
   if(is.character(kernel)) kernel <- match2DkernelName(kernel)
   isgauss <- identical(kernel, "gaussian")
@@ -435,10 +695,10 @@ smoothpointsEngine <- function(x, values, sigma, ...,
   
   if(isgauss &&
      spatstat.options("densityTransform") &&
-     spatstat.options("densityC")) {
+     spatstat.options("densityC") && !shrinking) {
     ## .................. experimental C code .....................
     if(debug)
-      cat('Using experimental code!\n')
+      cat('Transforming to standard coordinates (densityTransform=TRUE).\n')
     npts <- npoints(x)
     result <- numeric(npts)
     ## transform to standard coordinates
@@ -477,7 +737,7 @@ smoothpointsEngine <- function(x, values, sigma, ...,
                PACKAGE="spatstat.explore")
       if(sorted) result <- zz$result else result[oo] <- zz$result
     } else {
-      wtsort <- weights[oo]
+      wtsort <- if(sorted) weights else weights[oo]
       zz <- .C(SE_Gwtsmoopt,
                nxy     = as.integer(npts),
                x       = as.double(xx),
@@ -495,10 +755,10 @@ smoothpointsEngine <- function(x, values, sigma, ...,
       # Use mark of nearest neighbour (by l'Hopital's rule)
       result[nbg] <- values[nnwhich(x)[nbg]]
     }
-  } else if(isgauss && spatstat.options("densityC")) {
+  } else if(isgauss && spatstat.options("densityC") && !shrinking) {
     # .................. C code ...........................
     if(debug)
-      cat('Using standard code.\n')
+      cat('Using standard code (densityC=TRUE).\n')
     npts <- npoints(x)
     result <- numeric(npts)
     # sort into increasing order of x coordinate (required by C code)
@@ -527,7 +787,7 @@ smoothpointsEngine <- function(x, values, sigma, ...,
                  PACKAGE="spatstat.explore")
         if(sorted) result <- zz$result else result[oo] <- zz$result
       } else {
-        wtsort <- weights[oo]
+        wtsort <- if(sorted) weights else weights[oo]
         zz <- .C(SE_wtsmoopt,
                  nxy     = as.integer(npts),
                  x       = as.double(xx),
@@ -558,7 +818,7 @@ smoothpointsEngine <- function(x, values, sigma, ...,
                  PACKAGE="spatstat.explore")
         if(sorted) result <- zz$result else result[oo] <- zz$result
       } else {
-        wtsort <- weights[oo]
+        wtsort <- if(sorted) weights else weights[oo]
         zz <- .C(SE_awtsmoopt,
                  nxy     = as.integer(npts),
                  x       = as.double(xx),
@@ -581,6 +841,8 @@ smoothpointsEngine <- function(x, values, sigma, ...,
   } else {
     #' Either a non-Gaussian kernel or using older, partly interpreted code
     #' compute weighted densities
+    if(debug)
+      cat('Using partly-interpreted code.\n')
     if(is.null(weights)) {
       # weights are implicitly equal to 1
       numerator <- do.call(density.ppp,
@@ -634,6 +896,11 @@ smoothpointsEngine <- function(x, values, sigma, ...,
                                               list(...),
                                               list(edge=FALSE)))
     }
+    if(shrinking) {
+      ## shrinkage estimate
+      numerator <- shrinknumer + numerator
+      denominator <- shrinkdenom + denominator
+    }
     if(is.null(uhoh <- attr(numerator, "warnings"))) {
       result <- numerator/denominator
       result <- ifelseXB(is.finite(result), result, NA_real_)
@@ -670,34 +937,74 @@ markvar  <- function(X, sigma=NULL, ..., weights=NULL, varcov=NULL) {
   return(V)
 }
 
-bw.smoothppp <- function(X, nh=spatstat.options("n.bandwidth"),
+bw.smoothppp <- function(X, ...,
+                         nh=spatstat.options("n.bandwidth"),
                          hmin=NULL, hmax=NULL, warn=TRUE,
-                         kernel="gaussian") {
+                         kernel="gaussian", varcov1=NULL,
+                         train=NULL, test=NULL) {
   stopifnot(is.ppp(X))
   stopifnot(is.marked(X))
+  if(!is.null(varcov1))
+    check.nmatrix(varcov1, 2, things="spatial dimensions", mname="varcov1")
   if(is.function(kernel))
     stop("Custom kernel functions are not yet supported in bw.smoothppp")
   X <- coerce.marks.numeric(X)
-  # rearrange in ascending order of x-coordinate (for C code)
-  X <- X[fave.order(X$x)]
-  #
-  marx <- marks(X)
-  dimmarx <- dim(marx)
-  if(!is.null(dimmarx))
-    marx <- as.matrix(as.data.frame(marx))
-  # determine a range of bandwidth values
-#  n <- npoints(X)
+  if(!is.null(train)) {
+    ## training subset (logical vector)
+    train <- ppsubset(X, I=train, Iname='train', fatal=TRUE)
+  }
+  if(!is.null(test)) {
+    ## subset for prediction (logical vector)
+    test <- ppsubset(X, I=test, Iname='test', fatal=TRUE)
+  }
+  ## rearrange in ascending order of x-coordinate (for C code)
+  ord <- fave.order(X$x)
+  X <- X[ord]
+  if(!is.null(train)) train <- train[ord]
+  if(!is.null(test)) test <- test[ord]
+  #' sub-patterns
+  Xtrain <- if(is.null(train)) X else X[train]
+  Xtest  <- if(is.null(test))  X else X[test]
+  #' observed values 
+  ytrain <- marks(Xtrain)
+  ytest  <- marks(Xtest)
+  if(multicolumn <- !is.null(dim(ytrain))) {
+    ytrain <- as.matrix(as.data.frame(ytrain))
+    ytest  <- as.matrix(as.data.frame(ytest))
+  }
+  trainIndicator <- if(is.null(train)) NULL else as.numeric(train)
+  #' determine a range of bandwidth values
   if(is.null(hmin) || is.null(hmax)) {
     W <- Window(X)
-#    a <- area(W)
     d <- diameter(as.rectangle(W))
     # Stoyan's rule of thumb 
     stoyan <- bw.stoyan(X)
-    # rule of thumb based on nearest-neighbour distances
-    nnd <- nndist(X)
-    nnd <- nnd[nnd > 0]
+    #' rule of thumb based on nearest-neighbour distances
+    if(classic <- (is.null(test) && is.null(train))) {
+      nnd <- nndist(unique(X))
+    } else {
+      seqX <- seq_len(npoints(X))
+      itrain <- if(is.null(train)) seqX else which(train)
+      itest  <- if(is.null(test))  seqX else which(test)
+      nnd <- nncross(Xtrain, Xtest,
+                     iX=itrain, iY=itest,
+                     what="dist")
+    }
+    if(any(ok <- is.finite(nnd) & (nnd > 0))) {
+      nnd <- nnd[ok]
+    } else {
+      nnd <- d/16
+    }
+    if(!is.null(varcov1)) {
+      dref <- (det(varcov1))^(1/4)
+      d      <- d/dref
+      stoyan <- stoyan/dref
+      nnd    <- nnd/dref
+    }
     if(is.null(hmin)) {
-      hmin <- max(1.1 * min(nnd), stoyan/5)
+      hmin <- max(1.1 * min(nnd),
+                  stoyan/5,
+                  if(classic) 0 else quantile(nnd, 0.25))
       hmin <- min(d/8, hmin)
     }
     if(is.null(hmax)) {
@@ -705,17 +1012,29 @@ bw.smoothppp <- function(X, nh=spatstat.options("n.bandwidth"),
       hmax <- min(d/2, hmax)
     }
   } else stopifnot(hmin < hmax)
-  #
+  #' sequence of bandwidth values
   h <- geomseq(from=hmin, to=hmax, length.out=nh)
   cv <- numeric(nh)
-  # 
-  # compute cross-validation criterion
+  #' compute cross-validation criterion
   for(i in seq_len(nh)) {
-    yhat <- Smooth(X, sigma=h[i], at="points", leaveoneout=TRUE,
-                   kernel=kernel, sorted=TRUE)
-    if(!is.null(dimmarx))
+    #' Initially predict value at all locations
+    if(is.null(varcov1)) {
+      yhat <- Smooth(X, sigma = h[i],
+                     weights=trainIndicator,
+                     at="points", leaveoneout=TRUE,
+                     kernel=kernel, sorted=TRUE)
+    } else {
+      yhat <- Smooth(X, varcov = (h[i]^2)  * varcov1,
+                     weights=trainIndicator,
+                     at="points", leaveoneout=TRUE,
+                     kernel=kernel, sorted=TRUE)
+    }
+    if(multicolumn)
       yhat <- as.matrix(as.data.frame(yhat))
-    cv[i] <- mean((marx - yhat)^2)
+    #' Now restrict to test locations only, if required
+    if(!is.null(test))
+      yhat <- if(multicolumn) yhat[test,] else yhat[test]
+    cv[i] <- mean((ytest - yhat)^2)
   }
 
   # optimize
@@ -725,7 +1044,9 @@ bw.smoothppp <- function(X, nh=spatstat.options("n.bandwidth"),
                      criterion="Least Squares Cross-Validation",
                      warnextreme=warn,
                      hargnames=c("hmin", "hmax"),
-                     unitname=unitname(X))
+                     unitname=if(is.null(varcov1)) unitname(X) else NULL,
+                     template=varcov1, exponent=2,
+                     train=train, test=test)
   return(result)
 }
 
@@ -739,7 +1060,9 @@ smoothcrossEngine <- function(Xdata, Xquery, values, sigma, ...,
   validate2Dkernel(kernel)
   if(is.character(kernel)) kernel <- match2DkernelName(kernel)
   isgauss <- identical(kernel, "gaussian") && scalekernel
-
+  if(isTRUE(list(...)$se))
+    warning("Standard errors are not yet supported", call.=FALSE)
+  
   if(!is.null(dim(weights)))
     stop("weights must be a vector")
 
@@ -769,15 +1092,15 @@ smoothcrossEngine <- function(Xdata, Xquery, values, sigma, ...,
   if(bandwidth.is.infinite(sigma)) {
     #' uniform estimate
     if(is.null(weights)) weights <- rep(1, ndata)
-    single <- is.null(dim(values))
+    univariate <- is.null(dim(values))
     wtval <- weights * values 
     totwt <- sum(weights)
-    totwtval <- if(single) sum(wtval) else colSums(wtval)
+    totwtval <- if(univariate) sum(wtval) else colSums(wtval)
     denominator <- rep(totwt, nquery)
     numerator <- rep(totwtval, each=nquery)
-    if(!single) numerator <- matrix(numerator, nrow=nquery)
+    if(!univariate) numerator <- matrix(numerator, nrow=nquery)
     result <- numerator/denominator
-    if(!single)
+    if(!univariate)
       colnames(result) <- colnames(values)
     return(result)
   }
@@ -990,7 +1313,10 @@ smoothcrossEngine <- function(Xdata, Xquery, values, sigma, ...,
   return(result)
 }
 
-ExpSmoothLog <- function(X, ..., at=c("pixels", "points"), weights=NULL) {
+ExpSmoothLog <- function(X, ..., at=c("pixels", "points"), weights=NULL,
+                         se=FALSE) {
+  if(se)
+    stop("Standard errors are not yet supported when geometric=TRUE")
   verifyclass(X, "ppp")
   at <- match.arg(at)
   if(!is.null(weights)) 
@@ -1038,3 +1364,28 @@ ExpSmoothLog <- function(X, ..., at=c("pixels", "points"), weights=NULL) {
   }
   return(exp(Z))
 }
+
+VarOfWtdMean <- function(marx, weights) {
+  ## weighted average
+  totwt     <- sum(weights)
+  totwtmark <- sum(weights * marx)
+  Estimate  <- totwtmark/totwt
+  ## delta method approximation to variance of weighted average
+  varnum    <- sum(weights * marx^2)
+  varden    <- totwt
+  covnumden <- totwtmark
+  V <- DeltaMethodVarOfRatio(totwtmark, totwt, varnum, varden, covnumden)
+  return(V)
+}
+
+DeltaMethodVarOfRatio <- function(num, den, varnum, varden, covnumden,
+                                  positive=TRUE) {
+  Estimate <- num/den
+  V <- Estimate^2 * (
+    varnum/num^2
+    - 2 * covnumden/(num * den)
+    + varden/den^2
+  )
+  return(if(positive) posify(V) else V)
+}
+
